/*
McCreight's algorithm for suffix trees
*/

import java.util.*;

public class SuffixTree {
	String string; //input string
	int scan; //current scan pointer in input string
	Node token; //current position of the token in the suffix tree
	Node root; //root of the suffix tree
	Vector edges = new Vector(); //all edges in the suffix tree
	
	//main driver
	//read input string from the command line,
	//create suffix tree, then print it out
	//example call: suffixtree abracadabra$
	public static void main(String[] args) {
		String input = "";
		if (args.length > 0) input = args[0];
		try {
			SuffixTree st = new SuffixTree(input); //create the suffix tree
			System.out.println(Node.nextName + " nodes"); //print # of nodes
			for (int i = 0; i < st.edges.size(); i++) { //print edges
				System.out.println(st.edges.elementAt(i).toString());
			}
		} catch (IllegalStringException e) {
			System.out.println(e.getMessage());
		}
	}
	
	SuffixTree(String s) throws IllegalStringException {
		if (s.length() == 0) throw new IllegalStringException("Null string not allowed");
		//check that last character is the only occurrence of that character
		if (s.indexOf(s.charAt(s.length() - 1)) < s.length() - 1)
			throw new IllegalStringException("Last character must be unique");
		string = s; //set input string
		root = new Node(Node.INTERNAL); //root has no inEdge, is internal node
		root.mps = root; //max proper suffix of root is itself
		token = root; //start the token at the root
		scan = 0; //initialize scan pointer--index into input string
		for (int suffix = 0; suffix < string.length(); suffix++) { //for each suffix
			if (scan < suffix) scan = suffix; //update scan pointer if stale
			layIn(suffix); //lay in next suffix
		}
	}
	
	//lay in one suffix
	void layIn(int suffix) {
		Edge oldEdge, e = null;
		Node oldToken;
		int i;
		//find portion already laid in & move token
		if (token.mps != null) {
			token = token.mps;
		} else {
			oldToken = token;
			oldEdge = token.inEdge;
			i =  oldEdge.start;
			if (token.inEdge.source == root) i++; //if root, start with 2nd char
			token = token.inEdge.source.mps; //mps of parent node must exist
			while (i < oldEdge.end) { //if still in old edge, keep going
				e = findEdge(string.charAt(i),token); //find matching edge from token position
				if (e == null) throw new Error(); //should always be there
				i += e.end - e.start; //advance old edge pointer
				if (i <= oldEdge.end) token = e.sink; //if still in old edge, advance token
			}
			if (i > oldEdge.end) { //if new edge extends beyond old
				splitEdge(e,e.end + oldEdge.end - i); //split new edge
			}
			oldToken.mps = token; //update mps of old token location
		}
		//scan input string starting from the token
		while ((e = findEdge(string.charAt(scan),token)) != null) {
			int k = e.start;
			while (k < e.end && string.charAt(scan) == string.charAt(k)) {
				scan++; //scan down tree as far as possible
				k++;
			}
			if (k < e.end) { //if in the middle of an edge
				splitEdge(e,k); //split edge here
				break;
			}
			else token = e.sink; //else advance token
		}
		//put remainder of suffix into a new single edge
		new Edge(this,scan,string.length(),new Node(suffix));
	}
	
	//split an edge into 2 edges at point k
	void splitEdge(Edge e,int k) {
		token = new Node(e,Node.INTERNAL); //create new internal node
		new Edge(this,k,e.end,e.sink); //create new edge--goes AFTER old edge
		e.end = k; //update old edge data
		e.sink = token;
	}
	
	//find edge out of node n matching character c
	//return null if none exists	
	Edge findEdge(char c,Node n) {
		for (int j = 0; j < n.edges.size(); j++) {
			Edge e = (Edge)n.edges.elementAt(j);
			if (c == string.charAt(e.start)) return e;
		}
		return null; //not found
	}
}

//node of suffix tree
class Node {
	int name;
	static int nextName = 0; //name of node
	Edge inEdge = null; //unique incoming edge
	Node mps = null; //link to maximal proper suffix
	Vector edges = new Vector(); //outgoing edges
	static final int INTERNAL = -1; //indicates internal node
	int suffix = INTERNAL; //fill in with suffix # if a leaf
	
	//call with this constructor if inEdge does not exist
	Node(int s) {
		name = nextName++; //give it a new name
		suffix = s; //set suffix number
	}
	
	//call with this constructor if inEdge already exists
	Node(Edge e,int s) {
		this(s);
		inEdge = e; //incoming edge from parent
		//sink of inEdge is set in Edge constructor 
	}
}

class Edge {
	SuffixTree st; //suffix tree this edge belongs to
	int start; //index of first char in edge label
	int end; //index of last char + 1
	Node source = null; //edge points from source to sink
	Node sink = null;
	
	Edge(SuffixTree st,int start,int end,Node sink) {
		this.st = st; //suffix tree this edge appears in
		this.start = start; //index into input string of 1st char of label
		this.end = end; //index into input string of last char of label + 1
		this.sink = sink; //sink of this edge
		source = st.token; //source is always current token location
		source.edges.addElement(this); //add to list of edges of source
		st.edges.addElement(this); //add to list of edges of tree
		sink.inEdge = this; //set sink's inEdge to this edge
	}
	
	//construct a string representation of the edge
	//it is of the form:
	//a->b-->c [i,j) s , where
	//a is the source node,
	//b is the sink node,
	//c is the max proper suffix pointer (* if null)
	//[i,j) is the interval of the string label of this edge
	//s is the string label of this edge, which is the substring
	//of the input string corresponding to the interval [i,j).
	public String toString() {
		String s = source.name + "->" + sink.name + "-->";
		s += (sink.mps == null)?"*":String.valueOf(sink.mps.name);
		s += " [" + start + "," + end + ") ";
		s += st.string.substring(start,end);
		return s;
	}
}

class IllegalStringException extends Exception {
	IllegalStringException(String s) {
		super(s);
	}
}
